////////////////////////////////////////////////////////////////
//
// locproblem.cc
// Copyright (c) 2002 by Wesley H. Huang.  All rights reserved.
// 
////////////////////////////////////////////////////////////////
#include "locproblem.h"
#include <iostream>
#include <filereader.h>
#include <string>
#include <math.h>

//compensate for windows deficiency
#ifndef M_PI
#define M_PI 3.1415926535897932384626433832
#endif

LocProblem::LocProblem(const char* fname, bool verbose)
{
  string worldFile, robotFile, commandFile;
  double cdsr, casr, rsr;
  string sensing;

  const int MAX_PROBLEM_NAME_LEN = 100;
  char pn[MAX_PROBLEM_NAME_LEN];

  filereader::FileReader f;
  f.registerCString("ProblemName", pn, MAX_PROBLEM_NAME_LEN);
  f.registerScalar("World", &worldFile);
  f.registerScalar("Robot", &robotFile);
  f.registerScalar("Commands", &commandFile);
  f.registerVector3("start", &startX, &startY, &startAng);
  
  f.registerScalar("CommandDstSigmaRatio", &cdsr);
  f.registerScalar("CommandAngSigmaRatio", &casr);
  f.registerScalar("RangeSigmaRatio", &rsr);

  f.registerScalar("UseSensing", &sensing);

  f.readFile(fname);
  if (!f.complete()) {
    cerr << "ERROR: The following parameters were missing from the file..." 
	 << endl;
    list<string> m = f.missingParameters();
    for (list<string>::iterator k = m.begin();
	 k != m.end();
	 k++) 
      cerr << "  " << *k << endl;
    cerr << endl;
    exit(1);
  }

  problemName = string(pn);
  startAng *= M_PI/180.0;

  w = new World(worldFile.c_str(), verbose);
  r = new SimRobot(robotFile.c_str(), startX, startY, startAng,
		   w, cdsr, casr, rsr, verbose);
  c = new CommandList(commandFile.c_str(), (sensing == "yes"), verbose);
}


// read and store the commands from a file.  if "useSensing" is false
// then ignore any "sense" commands read
CommandList::CommandList(const char* fname, bool useSensing, 
			 bool verbose)
{
  if (verbose) cout << endl << "Opening file \"" << fname << "\"..." << endl;
  std::ifstream inpFile(fname);

  if (verbose) cout << "Reading commands..." << endl;
  while (true) {
    filereader::skipWhiteComment(inpFile);
    if (inpFile.eof()) break;
    Command c;
    inpFile >> c;
    Cmds.push_back(c);
  }
  if (verbose) cout << "End of file reached." << endl;

  current = Cmds.begin();
}
  
Command CommandList::getCommand() const 
{ 
  // work around windows const iterator == problem
  vector<Command>::const_iterator const_current = current;
  if (const_current == Cmds.end())
    return Command();
  else 
    return *current;
}

bool CommandList::commandsEnd() const
{ 
  // work around windows const iterator == problem
  vector<Command>::const_iterator const_current = current;

  return const_current == Cmds.end(); 
}


// read a Command from a stream
istream& operator>> (istream& in, Command &c)
{
  string ctype;
  in >> ctype;

  if (ctype == "move") {
    c.type = Command::move;
    in >> c.dist >> c.ang;
    c.ang *= M_PI/180.0;
  }
  else if (ctype == "sense") {
    c.type = Command::sense;
    in >> c.ang;
    c.dist = 0.0;
    c.ang *= M_PI/180.0;
  }
  else 
    cerr << "Unknown command type \"" << ctype 
	 << "\" in istream& operator>> (istream& in, Command& c)" << endl;

  return in;
}

// write a Command to a stream
//
// angle is printed in degrees
ostream& operator<< (ostream& out, const Command &c)
{
  if (c.type == Command::move) {
    out << "move: dist=" <<  c.dist << " ang=" << c.ang*180/M_PI;
  }
  else if (c.type == Command::sense) {
    out << "sense: ang=" << c.ang*180/M_PI;
  }
  else 
    cerr << "Unknown command type \"" << c.type 
	 << "\" in ostream& operator<< (ostream& out, const Command& c)"
	 << endl;

  return out;
}
