////////////////////////////////////////////////////////////////
//
// locproblem.h
// Copyright (c) 2002 by Wesley H. Huang.  All rights reserved.
// 
////////////////////////////////////////////////////////////////
#ifndef LOCPROBLEM_H
#define LOCPROBLEM_H

#include "world.h"
#include "simrobot.h"
#include <vector>
#include <iostream>

class CommandList;

class LocProblem {
public:
  string problemName;
  World *w;
  SimRobot *r;
  CommandList *c;
  
  double startX, startY;
  double startAng; // in radians

  // read in the problem description file
  LocProblem(const char* fname, bool verbose = false);
};

// a single command.  In a "move" command, "ang" represents the angle
// for a turn.  In a "sense" command, "ang" represents the direction
// the range sensor.  In both cases, "ang" is relative to the robot's
// heading.
// 
class Command {
public:
  enum CommandType {move, sense};
  CommandType type;
  double dist; 			// distance for a "move" command
  double ang;  // for "move" command, angle of turn (in radians)
               // for "sense" command, sensor heading (in radians)
               //   relative to robot heading

  // constructor for a move command
  Command(double d, double a) : type(Command::move), dist(d), ang(a) {}
  // constructor for a sense command
  Command(double a) : type(Command::sense), dist(0.0), ang(a) {}
  // default constructon
  Command() : dist(0.0), ang(0.0) {}
};

// read/write a Command from/to a stream
istream& operator>> (istream& in, Command& c);
ostream& operator<< (ostream& out, const Command& c);

// a class to manage the list of commands for the robot
//
// 
class CommandList {
private:
  vector<Command> Cmds;
  vector<Command>::iterator current;

public:
  // read command file and set current command to be the first
  //
  // if "useSensing" is false, then sensing commands are ignored when
  // reading the file.
  CommandList(const char* fname, bool useSensing, bool verbose = false);

  // methods to access commands by index (zero indexed)
  int numCommands() const { return Cmds.size();}
  Command getCommand(int i) const { return Cmds[i];}

  // methods to access commands sequentially
  Command getCommand() const; 
  Command nextCommand() { return *(current++); }
  Command prevCommand() { return *(current--); }
  bool commandsEnd() const;
  void restartCommands() { current = Cmds.begin(); }
};
  






#endif
